/// <reference types="ws" />
/// <reference types="node" />
import 'colors';
import * as WebSocket from 'isomorphic-ws';
export declare type Message = {
    command: string;
    id?: number;
    data?: any;
    reason?: any;
    socketId?: string | null;
    resumeToken?: string | null;
    packets?: Message[];
    resume?: string;
    items?: Message[];
};
/**
 * Ribbon - blazing-fast, msgpacked resumable WebSockets
 */
export declare class RibbonManager {
    /**
     * The number of reconnects that have been tried for the current session
     */
    reconnectCount: number;
    lastReconnect: number;
    extraPenalty: number;
    ws?: WebSocket;
    /**
     * The ID of the Ribbon
     */
    id?: string;
    /**
     * Resume ID for Ribbon, used for continuity after disconnects
     */
    resume?: string;
    /**
     * The ID of the last sent packet
     */
    lastSentId: number;
    /**
     * The ID of the last received packet
     */
    lastReceivedId: number;
    /**
     * List of the last 200 messages
     */
    lastSent: Message[];
    /**
     * Is the connection considered alive?
     */
    alive: boolean;
    /**
     * Reason the ribbon has been closed.
     */
    closeReason: string;
    /**
     * Object containing all of the current message listeners.
     */
    messageListeners: Record<string, ((data: any) => void)[]>;
    /**
     * Array of all onopen() listeners
     */
    openListeners: (() => void)[];
    /**
     * Array of all onclose() listeners
     */
    closeListeners: ((closeReason: string) => void)[];
    /**
     * Array of all onpong() listeners
     */
    pongListeners: ((delta: number) => void)[];
    /**
     * Array of all onresume() listeners
     */
    resumeListeners: ((arg0: number) => void)[];
    /**
     * Current pingInterval ID
     */
    pingInterval?: NodeJS.Timeout;
    /**
     * Time of the last ping, in ms
     */
    lastPing: number;
    /**
     * Will attempt to reconnect
     */
    mayReconnect: boolean;
    /**
     * Connection is dead
     */
    dead: boolean;
    ignoreCleanness: boolean;
    /**
     * We've pinged out; trying to reestablish connection
     */
    pingIssues: boolean;
    /**
     * It has connected at least once in the RibbonManager's lifetime
     */
    wasEverConnected: boolean;
    ribbonSessionId: string;
    /**
     * List of all Messages that need to be processed
     */
    incomingQueue: Message[];
    /**
     * We're about to switch endpoints
     */
    switchingEndpoint: boolean;
    /**
     * List of packets that should be sent on request
     */
    batchQueue: Uint8Array[];
    /**
     * Current batchQueue interval ID
     */
    batchTimeout?: NodeJS.Timeout;
    /**
     * Current endpoint
     */
    endpoint: string;
    /**
     * Initialize a ribbon
     * @param uri WebSocket Endpoint to connect to
     */
    constructor(uri?: string);
    /**
     * Smartly decide which method to send a message with
     * @param packet The message you want to encode
     */
    SmartEncode(packet: Message): Uint8Array;
    /**
     * Decode a packet, automatically detecting which method to use
     * @param packet The binary data to be decoded, as a Uint8Array
     */
    SmartDecode(packet: Uint8Array): Message;
    /**
     * Connect to the WebSocket and create a Ribbon.
     *
     */
    open(): void;
    /**
     * Compute a message
     * @param msg The message to be handled
     */
    HandleMessage(msg: Message): void;
    /**
     * Add a message to the processing queue
     * @param msg The message to be added
     */
    EnqueueMessage(msg: Message): void;
    /**
     * Send a message
     * @param command The name of the command to be executed
     * @param data The `data` object of the message
     * @param batched Add message to the {@link batchQueue}
     */
    send(command: string, data: any, batched?: boolean): void;
    /**
     * Send all batched messages
     */
    FlushBatch(): void;
    /**
     * Close the connection and don't reopen
     * @param reason A reason to show the user
     * @param silent Don't trigger close listeners
     * @example
     * // will safely close ribbon on ctrl-c
     * process.on('SIGINT', () => {
     *   ribbon.close('process terminated');
     * });
     */
    close(reason?: string, silent?: boolean): void;
    /**
     * Close the connection
     * @param penalty Extra time to wait before reconnecting
     */
    cut(penalty?: number): void;
    /**
     * Reconnect to new endpoint
     */
    switchEndpoint(): void;
    /**
     * Lose the current Ribbon and attempt to create a new one.
     */
    reconnect(): void;
    /**
     * Finish disconnection process and close listeners
     * @param silent Don't trigger close listeners
     */
    Die(silent?: boolean): void;
    getEndpoint(): string;
    /**
     * Change current endpoint
     * @param uri New endpoint
     */
    setEndpoint(uri: string): void;
    getId(): string | undefined;
    isAlive(): boolean;
    onclose(l: (closeReason: string) => void): void;
    /**
     * Triggered when Ribbon is established
     */
    onopen(l: () => void): void;
    /**
     * Triggered after every 'pong' (server response to ping)
     */
    onPong(l: (delta: number) => void): void;
    /**
     * Triggered when Ribbon is broken and resumed
     */
    onresume(l: (delay: number) => void): void;
    /**
     * Create an event listener for a message
     */
    on(type: string, l: (data: Record<string, any>) => void): void;
    /**
     * Remove an event listener for a message
     */
    off(type: string, l: (data: Record<string, any>) => void): void;
}
export default RibbonManager;
